unit _Hook;

/////////////////////////////////////////////////////////////////////
//
// Hi-Track Version 2
// Copyright (c) 2000-2002 Dmitry Liman [2:461/79]
//
// http://hi-files.by.ru
// http://hi-files.narod.ru
//
/////////////////////////////////////////////////////////////////////

interface

uses Objects, SysUtils, MyLib, FtnAddr, MsgApi, UnixMask, RegExpr;

////////////////////////////////////////////////////////////////
//
// AttrMask
//
////////////////////////////////////////////////////////////////

type
  PAttrMask = ^TAttrMask;
  TAttrMask = object
    procedure Assign( const S: String );
    function  Match( A: DWORD ) : Boolean;
    function  Apply( A: DWORD ) : DWORD;
  private
    Raised : DWORD;
    Lowered: DWORD;
  end; { AttrMask }

////////////////////////////////////////////////////////////////
//
// Hook
//
////////////////////////////////////////////////////////////////

type
  PHook = ^THook;
  THook = object (TObject)
    function Eat( const S: String ) : Boolean; virtual;
    function Hit( M: PFtnMessage ) : Boolean; virtual;
  end; { THook }

////////////////////////////////////////////////////////////////
//
// Mask
//
////////////////////////////////////////////////////////////////

type
  PMask = ^TMask;
  TMask = object (THook)
    function Eat( const S: String ) : Boolean; virtual;
    function Hit( M: PFtnMessage ) : Boolean; virtual;
  private
    pat_from : UnixMask.TState;
    pat_to   : UnixMask.TState;
    pat_orig : TAddress;
    pat_dest : TAddress;
    pat_subj : UnixMask.TState;
    pat_attr : TAttrMask;
  end; { TMask }

////////////////////////////////////////////////////////////////
//
// NameHook
//
////////////////////////////////////////////////////////////////

type
  PNameHook = ^TNameHook;
  TNameHook = object (THook)
    function Eat( const S: String ) : Boolean; virtual;
    function Hit( M: PFtnMessage ) : Boolean; virtual;
  private
    pattern: String;
  end; { TNameHook }

////////////////////////////////////////////////////////////////
//
// Clause
//
////////////////////////////////////////////////////////////////

type
  TClauseRef = (cr_dummy, cr_from, cr_to, cr_subj, cr_orig, cr_dest, cr_older, cr_newer);

  PClause = ^TClause;
  TClause = object (THook)
    constructor Init( R: TClauseRef );
  protected
    Inverted: Boolean;
    WhatRef : TClauseRef;
  end; { TClause }

  PStringClause = ^TStringClause;
  TStringClause = object (TClause)
    function Eat( const S: String ) : Boolean; virtual;
    function Hit( M: PFtnMessage ) : Boolean; virtual;
  private
    Pattern: UnixMask.TState;
  end; { TStringClause }

  PAddrClause = ^TAddrClause;
  TAddrClause = object (TClause)
    function Eat( const S: String ) : Boolean; virtual;
    function Hit( M: PFtnMessage ) : Boolean; virtual;
  private
    Pattern: TAddress;
  end; { TAddrClause }

  PTimeClause = ^TTimeClause;
  TTimeClause = object (TClause)
    function Eat( const S: String ) : Boolean; virtual;
    function Hit( M: PFtnMessage ) : Boolean; virtual;
  private
    Age: TDateTime;
  end; { TTimeClause }

  PAttrClause = ^TAttrClause;
  TAttrClause = object (TClause)
    function Eat( const S: String ) : Boolean; virtual;
    function Hit( M: PFtnMessage ) : Boolean; virtual;
  private
    Pattern: TAttrMask;
  end; { TAttrClause }

  PGrepClause = ^TGrepClause;
  TGrepClause = object (TClause)
    destructor Done; virtual;
    function Eat( const S: String ) : Boolean; virtual;
    function Hit( M: PFtnMessage ) : Boolean; virtual;
  private
    Pattern: TRegExprEngine;
  end; { TGrepClause }

  PUnionClause = ^TUnionClause;
  TUnionClause = object (TClause)
    constructor Init;
    destructor Done; virtual;
    function Eat( const S: String ) : Boolean; virtual;
    function Hit( M: PFtnMessage ) : Boolean; virtual;
    procedure AddClause( Clause: PClause );
  private
    Clauses: PCollection;
  end; { TUnionClause }

  PHubClause = ^THubClause;
  THubClause = object (TClause)
    function Eat( const S: String ) : Boolean; virtual;
    function Hit( M: PFtnMessage ) : Boolean; virtual;
  private
    Hub: TAddress;
  end; { THubClause }

////////////////////////////////////////////////////////////////
//
// ExtMask
//
////////////////////////////////////////////////////////////////

type
  PExtMask = ^TExtMask;
  TExtMask = object (THook)
    constructor Init;
    destructor Done; virtual;
    function Eat( const S: String ): Boolean; virtual;
    function Hit( M: PFtnMessage ) : Boolean; virtual;
    procedure AddClause( Clause: PClause; const S: String );
  private
    Clauses: PCollection;
    Union  : PUnionClause;
  end; { TExtMask }

{ =================================================================== }

implementation

uses _Config;

////////////////////////////////////////////////////////////////
//
// Assign
//
////////////////////////////////////////////////////////////////

{ --------------------------------------------------------- }
{ Assign                                                    }
{ --------------------------------------------------------- }

procedure TAttrMask.Assign( const S: String );
var
  j   : Integer;
  item: String;
  bit : DWORD;
begin
  Raised  := 0;
  Lowered := 0;

  if (S = '') or (S = '*') then
    Exit;

  if Length(S) mod 4 <> 0 then
    raise Exception.Create( 'Bad attr set: "' + S + '"' )
  else
    for j := 1 to Length(S) div 4 do
    begin
      item := Copy( S, 1 + Pred(j) * 4, 4 );
      bit  := WhatsAttr( Copy(Item, 2, 3) );
      if (bit = 0) or ((item[1] <> '+') and (item[1] <> '-')) then
        raise Exception.Create( 'Bad attr: "' + item + '"' )
      else if item[1] = '+' then
        SetBit( Raised, Bit, True )
      else
        SetBit( Lowered, Bit, True );
    end;
end; { Assign }

{ --------------------------------------------------------- }
{ Match                                                     }
{ --------------------------------------------------------- }

function TAttrMask.Match( A: DWORD ) : Boolean;
begin
  Result := ((A and Raised) = Raised) and ((A and Lowered) = 0);
end; { Match }

{ --------------------------------------------------------- }
{ Apply                                                     }
{ --------------------------------------------------------- }

function TAttrMask.Apply( A: DWORD ) : DWORD;
begin
  Result := (A and not Lowered) or Raised;
end; { Apply }

////////////////////////////////////////////////////////////////
//
// Hook
//
////////////////////////////////////////////////////////////////

{ --------------------------------------------------------- }
{ Eat                                                       }
{ --------------------------------------------------------- }

function THook.Eat( const S: String ) : Boolean;
begin
  Abstract;
end; { Eat }

{ --------------------------------------------------------- }
{ Hit                                                       }
{ --------------------------------------------------------- }

function THook.Hit( M: PFtnMessage ) : Boolean;
begin
  Abstract;
end; { Hit }

////////////////////////////////////////////////////////////////
//
// Mask
//
// <from>[, <orig>[, <to>[, <dest>[, <subj> [, <attr> ]]]]]
//
////////////////////////////////////////////////////////////////

{ --------------------------------------------------------- }
{ Eat                                                       }
{ --------------------------------------------------------- }

function TMask.Eat( const S: String ) : Boolean;
const
  any = '*';
var
  j: Integer;
  item: String;
begin
  UnixMask.MakeAutomatOn( any, pat_from );
  UnixMask.MakeAutomatOn( any, pat_to );
  UnixMask.MakeAutomatOn( any, pat_subj );
  pat_orig.Assign( any );
  pat_dest.Assign( any );
  pat_attr.Assign( any );

  for j := 1 to 6 do
  begin
    item := Trim( ExtractWord(j, S, _COMMA) );
    if item = '' then
      break;
    case j of
      1: UnixMask.MakeAutomatOn( item, pat_from );
      2: pat_orig.Assign( item );
      3: UnixMask.MakeAutomatOn( item, pat_to );
      4: pat_dest.Assign( item );
      5: UnixMask.MakeAutomatOn( item, pat_subj );
      6: pat_attr.Assign( item );
    end;
  end;

  if not pat_orig.Valid or not pat_dest.Valid then
    raise Exception.Create( 'Bad orig/dest addr' );

  Result := True;

end; { Eat }

{ --------------------------------------------------------- }
{ Hit                                                       }
{ --------------------------------------------------------- }

function TMask.Hit( M: PFtnMessage ) : Boolean;
begin
  Result := UnixMask.IsAcceptable( M^.GetFrom, pat_from ) and
            UnixMask.IsAcceptable( M^.GetTo,   pat_to )   and
            UnixMask.IsAcceptable( M^.GetSubj, pat_subj ) and
            pat_orig.Match( M^.GetOrig )                  and
            pat_dest.Match( M^.GetDest )                  and
            pat_attr.Match( M^.GetAttr );
end; { Hit }

////////////////////////////////////////////////////////////////
//
// NameHook
//
// Name <name>
//
////////////////////////////////////////////////////////////////

{ --------------------------------------------------------- }
{ Eat                                                       }
{ --------------------------------------------------------- }

function TNameHook.Eat( const S: String ) : Boolean;
begin
  pattern := S;
end; { Eat }

{ --------------------------------------------------------- }
{ Hit                                                       }
{ --------------------------------------------------------- }

function TNameHook.Hit( M: PFtnMessage ) : Boolean;
begin
  Result := (cmpi(M^.GetFrom, pattern) = 0) or
            (cmpi(M^.GetTo, pattern) = 0);
end; { Hit }

////////////////////////////////////////////////////////////////
//
// Clause
//
////////////////////////////////////////////////////////////////

{ --------------------------------------------------------- }
{ Init                                                      }
{ --------------------------------------------------------- }

constructor TClause.Init( R: TClauseRef );
begin
  inherited Init;
  WhatRef := R;
end; { Init }

////////////////////////////////////////////////////////////////
//
// StringClause
//
////////////////////////////////////////////////////////////////

{ --------------------------------------------------------- }
{ Eat                                                       }
{ --------------------------------------------------------- }

function TStringClause.Eat( const S: String ) : Boolean;
var
  j: Integer;
begin
  Result := True;
  if (S <> '') and (S[1] = '!') then
  begin
    UnixMask.MakeAutomatOn( Copy( S, 2, Length(S) ), Pattern );
    Inverted := True;
  end
  else
    UnixMask.MakeAutomatOn( S, Pattern );
end; { Eat }

{ --------------------------------------------------------- }
{ Hit                                                       }
{ --------------------------------------------------------- }

function TStringClause.Hit( M: PFtnMessage ) : Boolean;
begin
  case WhatRef of
    cr_from: Result := UnixMask.IsAcceptable( M^.GetFrom, Pattern );
    cr_to  : Result := UnixMask.IsAcceptable( M^.GetTo,   Pattern );
    cr_subj: Result := UnixMask.IsAcceptable( M^.GetSubj, Pattern );
  end;
  Result := Result xor Inverted;
end; { Hit }

////////////////////////////////////////////////////////////////
//
// AddrClause
//
////////////////////////////////////////////////////////////////

{ --------------------------------------------------------- }
{ Eat                                                       }
{ --------------------------------------------------------- }

function TAddrClause.Eat( const S: String ) : Boolean;
begin
  Result := True;
  if (S <> '') and (S[1] = '!') then
  begin
    Inverted := True;
    Pattern.Assign( Copy(S, 2, Length(S)) );
  end
  else
    Pattern.Assign( S );

  if not Pattern.Valid then
    raise Exception.Create( 'Bad FTN-addr in orig/dest clause' );

end; { Eat }

{ --------------------------------------------------------- }
{ Hit                                                       }
{ --------------------------------------------------------- }

function TAddrClause.Hit( M: PFtnMessage ) : Boolean;
begin
  case WhatRef of
    cr_orig: Result := pattern.Match( M^.GetOrig );
    cr_dest: Result := pattern.Match( M^.GetDest );
  end;
  Result := Result xor Inverted;
end; { Hit }

////////////////////////////////////////////////////////////////
//
// TimeClause
//
////////////////////////////////////////////////////////////////

{ --------------------------------------------------------- }
{ Eat                                                       }
{ --------------------------------------------------------- }

function TTimeClause.Eat( const S: String ) : Boolean;
begin
  Result := True;
  Age := SysUtils.Date - StrToInt( S );
end; { Eat }

{ --------------------------------------------------------- }
{ Hit                                                       }
{ --------------------------------------------------------- }

function TTimeClause.Hit( M: PFtnMessage ) : Boolean;
begin
  case WhatRef of
    cr_older: Result := M^.GetDate < Age;
    cr_newer: Result := M^.GetDate > Age;
  end;
end; { Hit }

////////////////////////////////////////////////////////////////
//
// AttrClause
//
////////////////////////////////////////////////////////////////

{ --------------------------------------------------------- }
{ Eat                                                       }
{ --------------------------------------------------------- }

function TAttrClause.Eat( const S: String ) : Boolean;
begin
  Result := True;
  Pattern.Assign( S );
end; { Eat }

{ --------------------------------------------------------- }
{ Hit                                                       }
{ --------------------------------------------------------- }

function TAttrClause.Hit( M: PFtnMessage ) : Boolean;
begin
  Result := Pattern.Match( M^.GetAttr );
end; { Hit }

////////////////////////////////////////////////////////////////
//
// GrepClause
//
////////////////////////////////////////////////////////////////

{ --------------------------------------------------------- }
{ Done                                                      }
{ --------------------------------------------------------- }

destructor TGrepClause.Done;
begin
  DestroyRegExprEngine( Pattern );
  inherited Done;
end; { Done }

{ --------------------------------------------------------- }
{ Eat                                                       }
{ --------------------------------------------------------- }

function TGrepClause.Eat( const S: String ) : Boolean;
var
  P: PChar;
begin
  Result := True;
  if (S = '') or ((S[1] <> '"') and (S[1] <> '''')) then
    raise Exception.Create( 'Regular expression missed (or not quoted)' );
  P := StrPNew( ExtractQuoted(S, 1) );
  Pattern := GenerateRegExprEngine( P, [ref_singleline, ref_caseinsensitive] );
  StrDispose( P );
  if Pattern.DestroyList = nil then
    raise Exception.Create( 'Bad regular expression' );
end; { Eat }

{ --------------------------------------------------------- }
{ Hit                                                       }
{ --------------------------------------------------------- }

function TGrepClause.Hit( M: PFtnMessage ) : Boolean;

  function Matched( V: VString ) : Boolean; far;
  var
    P: PChar;
    j: Integer;
    n: Integer;
  begin
    P := StrPNew( V^ );
    if P <> nil then
    begin
      Result := RegExprPos( Pattern, P, j, n );
      StrDispose( P );
    end
    else
      Result := False;
  end; { Matched }

begin
  Result := M^.FirstThat( @Matched ) <> nil;
end; { Hit }

////////////////////////////////////////////////////////////////
//
// HubClause
//
////////////////////////////////////////////////////////////////

{ --------------------------------------------------------- }
{ Eat                                                       }
{ --------------------------------------------------------- }

function THubClause.Eat( const S: String ) : Boolean;
begin
  Result := True;
  Hub.Assign( S );
  if not Hub.Valid then
    raise Exception.Create( 'Invalid hub address: "' + S + '"' );
end; { Eat }

{ --------------------------------------------------------- }
{ Hit                                                       }
{ --------------------------------------------------------- }

function THubClause.Hit( M: PFtnMessage ) : Boolean;
var
  A: TAddress;
  H: TAddress;
begin
  case WhatRef of
    cr_orig: A := M^.GetOrig;
    cr_dest: A := M^.GetDest;
  end;
  Result := False;
  if Cfg.OpenNodelist and
     Cfg.Nodelist^.FindHub( A, H )
  then
    Result := CompAddr( Hub, H ) = 0
end; { Hit }

////////////////////////////////////////////////////////////////
//
// UnionClause
//
////////////////////////////////////////////////////////////////

{ --------------------------------------------------------- }
{ Init                                                      }
{ --------------------------------------------------------- }

constructor TUnionClause.Init;
begin
  inherited Init( cr_dummy );
  Clauses := New( PCollection, Init(10, 10) );
end; { Init }

{ --------------------------------------------------------- }
{ Done                                                      }
{ --------------------------------------------------------- }

destructor TUnionClause.Done;
begin
  Destroy( Clauses );
  inherited Done;
end; { Done }

{ --------------------------------------------------------- }
{ Eat                                                       }
{ --------------------------------------------------------- }

function TUnionClause.Eat( const S: String ) : Boolean;
begin
  Result := True;
  if S = '' then
    Exit
  else if cmpi( S, 'Inverted' ) = 0 then
    Inverted := True
  else
    raise Exception.Create( 'Invalid "Union" clause' );
end; { Eat }

{ --------------------------------------------------------- }
{ Hit                                                       }
{ --------------------------------------------------------- }

function TUnionClause.Hit( M: PFtnMessage ) : Boolean;

  function MatchedClause( Clause: PClause ) : Boolean; far;
  begin
    Result := Clause^.Hit( M );
  end; { MatchedClause }

begin
  Result := (Clauses^.FirstThat( @MatchedClause ) <> nil) xor Inverted;
end; { Hit }

{ --------------------------------------------------------- }
{ AddClause                                                 }
{ --------------------------------------------------------- }

procedure TUnionClause.AddClause( Clause: PClause );
begin
  Clauses^.Insert( Clause );
end; { AddClause }

////////////////////////////////////////////////////////////////
//
// ExtMask
//
////////////////////////////////////////////////////////////////

{ --------------------------------------------------------- }
{ Init                                                      }
{ --------------------------------------------------------- }

constructor TExtMask.Init;
begin
  inherited Init;
  Clauses := New( PCollection, Init(10, 10) );
  Union   := nil;
end; { Init }

{ --------------------------------------------------------- }
{ Done                                                      }
{ --------------------------------------------------------- }

destructor TExtMask.Done;
begin
  Destroy( Clauses );
  inherited Done;
end; { Done }

{ --------------------------------------------------------- }
{ Eat                                                       }
{ --------------------------------------------------------- }

function TExtMask.Eat( const S: String ) : Boolean;
var
  Key : String;
  Val : String;
begin
  Result := False;
  SplitPair( S, Key, Val );
  if cmpi( Key, 'From' ) = 0 then
    AddClause( New(PStringClause, Init(cr_from)), Val )
  else if cmpi( Key, 'To' ) = 0 then
    AddClause( New( PStringClause, Init(cr_to)),  Val )
  else if cmpi( Key, 'Subj' ) = 0 then
    AddClause( New(PStringClause, Init(cr_subj)), Val )
  else if cmpi( Key, 'Orig' ) = 0 then
    AddClause( New(PAddrClause, Init(cr_orig)),   Val )
  else if cmpi( Key, 'Dest' ) = 0 then
    AddClause( New(PAddrClause, Init(cr_dest)),   Val )
  else if cmpi( Key, 'Older' ) = 0 then
    AddClause( New(PTimeClause, Init(cr_older)),  Val )
  else if cmpi( Key, 'Newer' ) = 0 then
    AddClause( New(PTimeClause, Init(cr_newer)),  Val )
  else if cmpi( Key, 'Attr' ) = 0 then
    AddClause( New(PAttrClause, Init(cr_dummy)),  Val )
  else if cmpi( Key, 'Grep' ) = 0 then
    AddClause( New(PGrepClause, Init(cr_dummy)),  Val )
  else if cmpi( Key, 'Orig_Hub' ) = 0 then
    AddClause( New(PHubClause, Init(cr_orig)),    Val )
  else if cmpi( Key, 'Dest_Hub' ) = 0 then
    AddClause( New(PHubClause, Init(cr_dest)),    Val )
  else if cmpi( Key, 'Hub' ) = 0 then
    AddClause( New(PHubClause, Init(cr_dest)),    Val )
  else if cmpi( Key, 'Union' ) = 0 then
  begin
    Union := New(PUnionClause, Init);
    Clauses^.Insert( Union );
    Union^.Eat( Val );
  end
  else if cmpi( Key, 'End' ) = 0 then
  begin
    if Union <> nil then
      Union := nil
    else
      Result := True;
  end
  else
    raise Exception.Create( 'Invalid clause: "' + S + '"' );
end; { Eat }

{ --------------------------------------------------------- }
{ AddClause                                                 }
{ --------------------------------------------------------- }

procedure TExtMask.AddClause( Clause: PClause; const S: String );
begin
  if Union <> nil then
    Union^.AddClause( Clause )
  else
    Clauses^.Insert( Clause );
  Clause^.Eat( S );
end; { AddClause }

{ --------------------------------------------------------- }
{ Hit                                                       }
{ --------------------------------------------------------- }

function TExtMask.Hit( M: PFtnMessage ) : Boolean;

  function DontMatch( Clause: PClause ) : Boolean; far;
  begin
    Result := not Clause^.Hit( M );
  end; { Hit }

begin
  Result := Clauses^.FirstThat( @DontMatch ) = nil;
end; { Hit }

end.



