unit _Target;

/////////////////////////////////////////////////////////////////////
//
// Hi-Track Version 2
// Copyright (c) 2000-2002 Dmitry Liman [2:461/79]
//
// http://hi-files.by.ru
// http://hi-files.narod.ru
//
/////////////////////////////////////////////////////////////////////

interface

uses
  Objects, _Catcher;

////////////////////////////////////////////////////////////////
//
// Target
//
////////////////////////////////////////////////////////////////

type
  PTarget = ^TTarget;
  TTarget = object (TCollection)
    constructor Init;
    destructor Done; virtual;
    procedure AddStorage( const S: String );
    procedure Shot;
    function  AddCatcher( Catcher: PCatcher ) : PCatcher;
  private
    Catchers: PCollection;
  end; { TTarget }

////////////////////////////////////////////////////////////////
//
// Storage
//
////////////////////////////////////////////////////////////////

type
  TMaintenanceOption = ( mo_Renumber, mo_Force, mo_Auto );
  TMaintOptSet = set of TMaintenanceOption;

  PStorage = ^TStorage;
  TStorage = object (TObject)
    constructor Init( T: PTarget );
    procedure Eat( const S: String ); virtual;
    procedure Shot; virtual;
  protected
    Options: TMaintOptSet;
    Path   : String;
    Target : PTarget;
  end; { TStorage }

  PMsgStorage = ^TMsgStorage;
  TMsgStorage = object (TStorage)
    procedure Shot; virtual;
  private
    procedure Maintenance;
    procedure Renumber;
  end; { TMsgStorage }

  PPktStorage = ^TPktStorage;
  TPktStorage = object (TStorage)
    procedure Shot; virtual;
  end; { TPktStorage }

  PBsoStorage = ^TBsoStorage;
  TBsoStorage = object (TStorage)
    procedure Shot; virtual;
  private
    procedure Maintenance;
    procedure Unpack;
    function  AutoDepend: Boolean;
  end; { TBsoStorage }

{ =================================================================== }

implementation

uses
  MyLib, SysUtils, _Preproc, MsgAPI, _Config, FtnAddr, _BSO, _Action;

////////////////////////////////////////////////////////////////
//
// IntVector
//
////////////////////////////////////////////////////////////////

type
  PIntVector = ^TIntVector;
  TIntVector = object (TSortedCollection)
    procedure FreeItem( Item: Pointer ); virtual;
    function Compare( Key1, Key2: Pointer ) : Integer; virtual;
    procedure Add( Item: Integer );
  end; { TIntVector }

{ --------------------------------------------------------- }
{ FreeItem                                                  }
{ --------------------------------------------------------- }

procedure TIntVector.FreeItem( Item: Pointer );
begin
end; { FreeItem }

{ --------------------------------------------------------- }
{ Compare                                                   }
{ --------------------------------------------------------- }

function TIntVector.Compare( Key1, Key2: Pointer ) : Integer;
begin
  if Integer(Key1) < Integer(Key2) then
    Result := -1
  else if Integer(Key1) > Integer(Key2) then
    Result := 1
  else
    Result := 0;
end; { Compare }

{ --------------------------------------------------------- }
{ Add                                                       }
{ --------------------------------------------------------- }

procedure TIntVector.Add( Item: Integer );
begin
  Insert( Pointer(Item) );
end; { Add }

////////////////////////////////////////////////////////////////
//
// TTarget
//
////////////////////////////////////////////////////////////////

{ --------------------------------------------------------- }
{ Init                                                      }
{ --------------------------------------------------------- }

constructor TTarget.Init;
begin
  inherited Init(10, 10);
  Catchers := New( PCollection, Init(10, 10) );
end; { Init }

{ --------------------------------------------------------- }
{ Done                                                      }
{ --------------------------------------------------------- }

destructor TTarget.Done;
begin
  Destroy( Catchers );
  inherited Done;
end; { Done }

{ --------------------------------------------------------- }
{ AddStorage                                                }
{ --------------------------------------------------------- }

procedure TTarget.AddStorage( const S: String );
var
  Key: String;
  Val: String;
  Storage: PStorage;
begin
  SplitPair( S, Key, Val );
  if cmpi( Key, 'MSG' ) = 0 then
  begin
    Storage := New( PMsgStorage, Init(@Self) );
    Insert( Storage );
    Storage^.Eat( Val );
  end
  else if cmpi( Key, 'PKT' ) = 0 then
  begin
    Storage := New( PPktStorage, Init(@Self) );
    Insert( Storage );
    Storage^.Eat( Val );
  end
  else if cmpi( Key, 'BSO' ) = 0 then
  begin
    Storage := New( PBsoStorage, Init(@Self) );
    Insert( Storage );
    Storage^.Eat( Val );
  end
  else
    raise Exception.Create( 'Unknown target storage: "' + Key + '"' );
end; { AddStorage }

{ --------------------------------------------------------- }
{ AddCatcher                                                }
{ --------------------------------------------------------- }

function TTarget.AddCatcher( Catcher: PCatcher ) : PCatcher;
begin
  Catchers^.Insert( Catcher );
  Result := Catcher;
end; { NewCatcher }

{ --------------------------------------------------------- }
{ Shot                                                      }
{ --------------------------------------------------------- }

procedure TTarget.Shot;
  procedure ShotStorage( S: PStorage ); far;
  begin
    S^.Shot;
  end;
begin
  ForEach( @ShotStorage );
end; { Shot }

////////////////////////////////////////////////////////////////
//
// Storage
//
////////////////////////////////////////////////////////////////

{ --------------------------------------------------------- }
{ Init                                                      }
{ --------------------------------------------------------- }

constructor TStorage.Init( T: PTarget );
begin
  inherited Init;
  Target := T;
end; { Init }

{ --------------------------------------------------------- }
{ Eat                                                       }
{ --------------------------------------------------------- }

procedure TStorage.Eat( const S: String );
var
  j  : Integer;
  opt: String;
begin
  for j := 1 to WordCount( S, _BLANK ) do
  begin
    opt := ExtractWord( j, S, _BLANK );
    if opt[1] <> '+' then
      Path := opt
    else
    begin
      if cmpi(opt, '+renumber') = 0 then
        Include(Options, mo_Renumber)
      else if cmpi(opt, '+auto') = 0 then
        Include(Options, mo_auto)
      else if cmpi(opt, '+force') = 0 then
        Include(Options, mo_force)
      else
        raise Exception.Create( 'Unknown maintenance option: "' + opt + '"' );
    end
  end;
  if Path <> '' then
    CheckDir( Path, 'Storage' );
end; { Eat }

{ --------------------------------------------------------- }
{ Shot                                                      }
{ --------------------------------------------------------- }

procedure TStorage.Shot;
begin
  Abstract;
end; { Shot }

////////////////////////////////////////////////////////////////
//
// MsgStorage
//
////////////////////////////////////////////////////////////////

{ --------------------------------------------------------- }
{ Shot                                                      }
{ --------------------------------------------------------- }

procedure TMsgStorage.Shot;

  procedure Attack( Name: VString ); far;
  var
    M: PNetmailMessage;
    Changed: Boolean;

    procedure PickupMessage( Catcher: PCatcher ); far;
    begin
      if Catcher^.Pickup( M ) then
        Changed := True;
    end; { PickupMessage }

  begin
    M := New( PNetmailMessage, Init );
    try
      Changed := False;
      M^.Load( AtPath(Name^, Path) );
      Target^.Catchers^.ForEach( @PickupMessage );
      if Changed then
        M^.Save;
    except
      on I: EIgnoreAction do
        Log( ll_Action, I.Message );
      on E: Exception do
        Log( ll_Error, AtPath(Name^, Path) + ' error: ' + E.Message );
    end;
    Destroy( M );
  end; { Attack }

var
  SR: TSearchRec;
  Files: PStrings;
  DosError: Integer;

begin
  if Path = '' then
    Path := Cfg.Netmail;

  Log( ll_Target, 'Processing MSG-base ' + Path );

  Maintenance;

  Files := New( PStrings, Init(1000, 1000) );
  DosError := SysUtils.FindFirst( AtPath('*.MSG', Path), faArchive, SR );
  while DosError = 0 do
  begin
    Files^.Insert( AllocStr(SR.Name) );
    DosError := SysUtils.FindNext( SR );
  end;
  SysUtils.FindClose( SR );

  try
    Files^.ForEach( @Attack );
  finally
    Log( ll_Target, IntToStr(Files^.Count) + ' messages processed' );
    Destroy( Files );
  end;

end; { Shot }

{ --------------------------------------------------------- }
{ Maintenance                                               }
{ --------------------------------------------------------- }

procedure TMsgStorage.Maintenance;
begin
  if mo_Renumber in Options then
    Renumber;
end; { Maintenance }

{ --------------------------------------------------------- }
{ Renumber                                                  }
{ --------------------------------------------------------- }

procedure TMsgStorage.Renumber;
var
  j : Integer;
  n : Integer;
  SR: TSearchRec;
  Files   : PIntVector;
  Bottom  : Integer;
  DosError: Integer;
begin
  Log( ll_Target, 'Maintenance: renumbering MSG-base...' );

  Files := New( PIntVector, Init(1000, 1000) );

  DosError := FindFirst( AtPath('*.MSG', Path), faArchive, SR );
  while DosError = 0 do
  begin
    try
      j := StrToInt( ChangeFileExt(SR.Name, '') );
      Files^.Add( j );
    except
      { ...nothing...}
    end;
    DosError := FindNext( SR );
  end;
  FindClose( SR );

  Bottom := 0;
  for j := 0 to Pred(Files^.Count) do
  begin
    Inc( Bottom );
    n := Integer( Files^.At(j) );
    if n > Bottom then
    begin
      if not RenameFile( AtPath( IntToStr(n) + '.MSG', Path ),
                        AtPath( IntToStr(Bottom) + '.MSG', Path ) )
      then
        Log( ll_Error, '? Renumber: failed to rename ' + IntToStr(n) + ' -> ' +
             IntToStr(Bottom) + ' at ' + Path );
    end;
  end;

  Log( ll_Target, IntToStr(Files^.Count) + ' MSG files renumbered' );
  Destroy( Files );

end; { Renumber }

////////////////////////////////////////////////////////////////
//
// PktStorage
//
////////////////////////////////////////////////////////////////

{ --------------------------------------------------------- }
{ Shot                                                      }
{ --------------------------------------------------------- }

procedure TPktStorage.Shot;

  procedure Attack( Name: VString ); far;
  var
    j: Integer;
    Pkt: PPacket;
    Msg: PPackedMessage;
    Changed: Boolean;

    procedure PickupMessage( Catcher: PCatcher ); far;
    begin
      if Catcher^.Pickup( Msg ) then
        Changed := True;
    end; { PickupMessage }

  begin
    Pkt := New( PPacket, Init );

    try
      Pkt.Load( AtPath(Name^, Path) );

      Changed := False;
      for j := 0 to Pred(Pkt^.Count) do
      begin
        Msg := Pkt^.At(j);
        try
          Target^.Catchers^.ForEach( @PickupMessage );
        except
          on I: EIgnoreAction do
            Log( ll_Action, I.Message );
        end;
      end;

      if Changed then
          Pkt^.Save;

    except
      on E: Exception do
        Log( ll_Error, Name^ + ' error: ' + E.Message );
    end;

    Destroy( Pkt );
  end; { Attack }

var
  SR: TSearchRec;
  Files: PStrings;
  DosError: Integer;

begin
  if Path = '' then
    Path := Cfg.Inbound;

  Log( ll_Target, 'Processing PKT-base ' + Path );

  Files := New( PStrings, Init(1000, 1000) );
  DosError := SysUtils.FindFirst( AtPath('*.PKT', Path), faArchive, SR );
  while DosError = 0 do
  begin
    Files^.Insert( AllocStr(SR.Name) );
    DosError := SysUtils.FindNext( SR );
  end;
  SysUtils.FindClose( SR );

  try
    Files^.ForEach( @Attack );
  finally
    Log( ll_Target, IntToStr(Files^.Count) + ' packets processed' );
    Destroy( Files );
  end;

end; { Shot }

////////////////////////////////////////////////////////////////
//
// BsoStorage
//
////////////////////////////////////////////////////////////////

{ --------------------------------------------------------- }
{ Unpack                                                    }
{ --------------------------------------------------------- }

procedure TBsoStorage.Unpack;
var
  SR: TSearchRec;
  Files: PStrings;
  DosError: Integer;
  FirstMsg: Boolean;

  procedure ScanPntDir( const Pnt: String );
  var
    SR: TSearchRec;
    DosError: Integer;
    S: String;
  begin
    S := AddBackSlash(Path) + Pnt + '\*.?UT';
    DosError := SysUtils.FindFirst( S, faArchive, SR );
    while DosError = 0 do
    begin
      Files^.Insert( AllocStr(Pnt + '\' + SR.Name) );
      DosError := SysUtils.FindNext( SR );
    end;
    SysUtils.FindClose( SR );
  end; { ScanPntDir }

  procedure StripOurVia( M: PFtnMessage );
  var
    j: Integer;
    Pattern: String;
  begin
    Pattern := ^A'Via ' + Cfg.MakeOurViaStamp;
    Pattern := Copy( Pattern, 1, Pos('@', Pattern) );
    for j := Pred(M^.Via^.Count) downto 0 do
    begin
      if Pos( Pattern, VString(M^.Via^.At(j))^ ) = 1 then
        M^.Via^.AtFree( j )
      else
        Exit;
    end;
  end; { StripOurVia }

  procedure UnpackPkt( const Name: String; A: TAddress );
  var
    Pkt: PPacket;

    procedure UnpackMessage( M: PPackedMessage ); far;
    var
      U: PNetmailMessage;
    begin
      try
        U := New( PNetmailMessage, Init );
        U^.Clone( M );
        StripOurVia( U );
        U^.SetAttr( U^.GetAttr and not ATTR_SENT );
        if FirstMsg then
        begin
          U^.SaveAs( NewMsgFileName(Cfg.Netmail) );
          FirstMsg := False;
        end
        else
          U^.SaveAs( NextMsgFileName(Cfg.Netmail) );
        M^.Kill;
      finally
        Destroy( U );
      end;
    end; { UnpackMessage }

  begin
    Log( ll_Target, 'Unpacking ' + ExtractFileExt(Name) + ' for ' + A.MakeStr );
    if LockAddr( A ) then
    begin
      try
        Pkt := New( PPacket, Init );
        Pkt^.Load( Name );
        Pkt^.ForEach( @UnpackMessage );
      except
        on E: Exception do
          Log( ll_Error, Name + ' error: ' + E.Message );
      end;
      Pkt^.Save;
      Destroy( Pkt );
      UnlockAddr( A );
    end
    else
      Log( ll_Target, 'Address ' + A.MakeStr + ' is busy now :(' );
  end; { UnpackPkt }

  procedure Attack( Name: VString ); far;
  var
    S: String;
    A: TAddress;
  begin
    A.Point := 0;
    if Pos('\', Name^) = 0 then
      S := Name^
    else
    begin
      // we have a point
      S := ChangeFileExt( ExtractFileName(Name^), '' );
      try
        A.Point := HexToInt( S );
      except
        Log( ll_Error, 'Bad file name in BSO, skipping ' + Name^ );
        Exit;
      end;
      S := Copy( Name^, 1, Pos('\', Name^) - 1 );
    end;
    S := ChangeFileExt( S, '' );
    try
      if Length(S) <> 8 then
        raise Exception.Create( '' );
      A.Net  := HexToInt( Copy(S, 1, 4) );
      A.Node := HexToInt( Copy(S, 5, 4) );
    except
      Log( ll_Error, 'Bad file name in BSO, skipping ' + Name^ );
      Exit;
    end;
    A.Zone := Cfg.PrimaryAddr.Zone;
    UnpackPkt( AddBackSlash(Path) + Name^, A );
  end; { Attack }

begin
  Files := New( PStrings, Init(1000, 1000) );
  DosError := SysUtils.FindFirst( AtPath('*.?UT', Path), faArchive, SR );
  while DosError = 0 do
  begin
    Files^.Insert( AllocStr(SR.Name) );
    DosError := SysUtils.FindNext( SR );
  end;
  SysUtils.FindClose( SR );

  DosError := SysUtils.FindFirst( AtPath('*.PNT', Path), faDirectory, SR );
  while DosError = 0 do
  begin
    ScanPntDir( SR.Name );
    DosError := SysUtils.FindNext( SR );
  end;
  SysUtils.FindClose( SR );

  try
    FirstMsg := True;
    Files^.ForEach( @Attack );
    Cfg.Session.BsoAge := SysUtils.Now;
    Cfg.SesChanged := True;
  finally
    Log( ll_Target, IntToStr(Files^.Count) + ' packets processed' );
    Destroy( Files );
  end;

end; { Unpack }

{ --------------------------------------------------------- }
{ Maintenance                                               }
{ --------------------------------------------------------- }

procedure TBsoStorage.Maintenance;
begin
  if mo_force in Options then
    Unpack
  else if (mo_auto in Options) and AutoDepend then
    Unpack;
end; { Maintenance }

{ --------------------------------------------------------- }
{ Shot                                                      }
{ --------------------------------------------------------- }

procedure TBsoStorage.Shot;
begin
  if Path = '' then
    Path := Cfg.Outbound;

  Log( ll_Target, 'Processing BSO ' + Path );

  Maintenance;
end; { Shot }

{ --------------------------------------------------------- }
{ AutoDepend                                                }
{ --------------------------------------------------------- }

function TBsoStorage.AutoDepend: Boolean;
begin
  Result := True;
  with Cfg.Session do
  begin
    if CfgAge > BsoAge then
      Log( ll_Target, 'BSO auto-dependency: configuration changed' )
    else if IdxAge > BsoAge then
      Log( ll_Target, 'BSO auto-dependency: nodelist index changed' )
    else
      Result := False;
  end;
end; { AutoDepend }

end.

