unit MsgApi;

/////////////////////////////////////////////////////////////////////
//
// Hi-Track Version 2
// Copyright (c) 2000-2002 Dmitry Liman [2:461/79]
//
// http://hi-files.by.ru
// http://hi-files.narod.ru
//
/////////////////////////////////////////////////////////////////////

interface

uses
  Objects, MyLib, FtnAddr, SysUtils;

const
   // FTS-0001 attribute flags
   ATTR_PRIVATE  = $0001;        // Private
   ATTR_CRASH    = $0002;        // Crash (high-priority mail)
   ATTR_RECVD    = $0004;        // Received
   ATTR_SENT     = $0008;        // Sent
   ATTR_FILEATT  = $0010;        // File attach
   ATTR_TRANSIT  = $0020;        // Transit
   ATTR_ORPHAN   = $0040;        // Orphan
   ATTR_KILLSENT = $0080;        // Kill/Sent
   ATTR_LOCAL    = $0100;        // Local
   ATTR_HOLD     = $0200;        // Hold
   ATTR_UNUSED   = $0400;        // Unused !!!
   ATTR_FREQ     = $0800;        // File request
   ATTR_RRQ      = $1000;        // Receipt request
   ATTR_RRC      = $2000;        // Is return receipt
   ATTR_ARQ      = $4000;        // Audit request
   ATTR_URQ      = $8000;        // Update request

   // Extended attribute flags (used in FLAGS kludge)
   ATTR_ARCSENT   = $00010000;   // Archive/Sent
   ATTR_DIR       = $00020000;   // Direct
   ATTR_ZONEGATE  = $00040000;   // Zonegate
   ATTR_HUB       = $00080000;   // Host/Hub routing
   ATTR_IMM       = $00100000;   // Immediate
   ATTR_XMAIL     = $00200000;   // Compressed mail attached
   ATTR_TRUNCSENT = $00400000;   // Truncate file sent
   ATTR_LOCK      = $00800000;   // Locked
   ATTR_CFM       = $01000000;   // Confirm request
   ATTR_FPU       = $02000000;   // Force pickup

function WhatsAttr( const S: String ) : DWORD;
function AttrStr( Attr: DWORD ) : String;
function MsgDateStr( D: TDateTime ) : String;
function is_netmail( const echotag: String ) : Boolean;

function NewPktFileName ( const Path: String ) : String;
function NewMsgFileName ( const Path: String ) : String;
function NextMsgFileName( const Path: String ) : String;

function MakePrintable( S: String ) : String;
function Invalidate( S: String ) : String;

const
  _NETMAIL  = 'Netmail';
  _ORIGIN   = ' * Origin: ';
  _TEARLINE = '--- ';

////////////////////////////////////////////////////////////////
//
// FtnMessage
//
////////////////////////////////////////////////////////////////

type
  PFtnMessage = ^TFtnMessage;
  TFtnMessage = object (TStrings)
    constructor Init;
    destructor Done; virtual;

    function GetArea  : String;
    function GetFrom  : String;
    function GetTo    : String;
    function GetOrig  : TAddress;
    function GetDest  : TAddress;
    function GetSubj  : String;
    function GetAttr  : DWORD;
    function GetFlags : String;
    function GetDate  : TDateTime;
    function GetMsgID : String;
    function GetReply : String;

    function HasOrigin   : Boolean;
    function GetOrigin   : String;
    function MakeOrigin  : String;
    function HasTearline : Boolean;
    function GetTearline : String;
    function MakeTearline: String;

    procedure SetArea( const S: String );
    procedure SetFrom( const S: String );
    procedure SetTo( const S: String );
    procedure SetOrig( const A: TAddress );
    procedure SetDest( const A: TAddress );
    procedure SetSubj( const S: String );
    procedure SetAttr( A: DWORD );
    procedure SetFlags( const S: String );
    procedure SetDate( D: TDateTime );
    procedure SetReply( const S: String );
    procedure SetOrigin( const S: String );
    procedure DropOrigin( const S: String );
    procedure SetTearline( const S: String );

    procedure AddVia( const S: String );

    function  Echomail: Boolean;
    function  Deleted: Boolean;
    procedure Kill; virtual;
    procedure Clone( Source: PFtnMessage );

    procedure BinSave( var Bin: TStream );
    procedure TxtSave( var Txt: Text );

  protected
    procedure UpdateMsgID;
    function  CheckKludge( const S: String ) : Boolean;
    procedure ParseFlags( const S: String );
    procedure SaveBinHeader( var Bin: TStream ); virtual;

  protected
    _From    : String;
    _To      : String;
    Orig     : TAddress;
    Dest     : TAddress;
    Subj     : String;
    Date     : TDateTime;
    Area     : String;
    MsgID    : String;
    Reply    : String;
    Flags    : String;
    Attr     : DWORD;
    Origin   : VString;
    Tearline : VString;
    Rip      : Boolean;

  public
    Path     : PStrings;
    Seenby   : PStrings;
    Via      : PStrings;

  end; { TFtnMessage }

////////////////////////////////////////////////////////////////
//
// NetmailMessage
//
////////////////////////////////////////////////////////////////

type
  PNetmailMessage = ^TNetmailMessage;
  TNetmailMessage = object (TFtnMessage)
    procedure Load( FileName: String );
    procedure Save;
    procedure SaveAs( const FileName: String );
    procedure Kill; virtual;
  protected
    procedure SaveBinHeader( var bin: TStream ); virtual;
  private
    MsgName: String;
  end; { TNetmailMessage }

////////////////////////////////////////////////////////////////
//
// PackedMessage
//
////////////////////////////////////////////////////////////////

type
  PPackedMessage = ^TPackedMessage;
  TPackedMessage = object (TFtnMessage)
{$IFDEF USE_FILE_MAPPING}
    procedure Load( var Pkt: PChar; PktOrig, PktDest: TAddress );
{$ENDIF}
  protected
    procedure SaveBinHeader( var bin: TStream ); virtual;
  end; { TPackedMessage }

////////////////////////////////////////////////////////////////
//
// Packet
//
////////////////////////////////////////////////////////////////

type
  PPacket = ^TPacket;
  TPacket = object (TCollection)
    constructor Init;
    procedure Load( FileName: String );
    procedure Save;
    procedure SaveAs( const FileName: String );

    procedure SetOrig( const A: TAddress );
    procedure SetDest( const A: TAddress );
    procedure SetDate( D: TDateTime );

  private
    PktName : String;
    Orig    : TAddress;
    Dest    : TAddress;
    Date    : TDateTime;
    Password: String;
  end; { TPacket }

{ =================================================================== }

implementation

{$IFDEF VIRTUALPASCAL}
uses
  VpSysLow,
  {$IFDEF USE_FILE_MAPPING}
  Windows;
  {$ENDIF}
{$ENDIF}

{$IFDEF FPC}
  {$IFDEF USE_FILE_MAPPING}
uses Windows;
  {$ENDIF}
{$ENDIF}

const
  STREAM_BUFFER_SIZE = 8 * 1024;

  ATTR_FLAGS_COUNT = 26;
  AttrFlag: array [0..Pred(ATTR_FLAGS_COUNT)] of String[3] = (
    'PVT', 'CRA', 'RCV', 'SNT', 'FIL', 'TRS', 'ORP', 'K/S',
    'LOC', 'HLD', 'UNU', 'FRQ', 'RRQ', 'RRC', 'ARQ', 'URQ',
    'A/S', 'DIR', 'ZON', 'HUB', 'IMM', 'XMA', 'TFS', 'LOK',
    'CFM', 'FPU' );

{ --------------------------------------------------------- }
{ WhatsAttr                                                 }
{ --------------------------------------------------------- }

function WhatsAttr( const S: String ) : DWORD;
var
  j: Integer;
begin
  for j := 0 to Pred(ATTR_FLAGS_COUNT) do
    if cmpi( S, AttrFlag[j] ) = 0 then
    begin
      Result := (1 shl j);
      Exit;
    end;
  Result := 0;
end; { WhatsAttr }

{ --------------------------------------------------------- }
{ AttrStr                                                   }
{ --------------------------------------------------------- }

function AttrStr( Attr: DWORD ) : String;
var
  j: Integer;
begin
  Result := '';
  for j := 0 to Pred(ATTR_FLAGS_COUNT) do
    if TestBit( Attr, 1 shl j ) then
      Result := Result + AttrFlag[j] + ' ';
  Result := TrimRight(Result);
end; { AttrStr }

{ --------------------------------------------------------- }
{ FlagsStr                                                  }
{ --------------------------------------------------------- }

function FlagsStr( Attr: DWORD; const Flags: String ) : String;
begin
  Result := AttrStr( Attr );
  if Result <> '' then
    Result := Result + ' ';
  Result := Result + Flags;
end; { FlagsStr }

{ --------------------------------------------------------- }
{ Strip_fts1_attr                                           }
{ --------------------------------------------------------- }

function Strip_fts1_attr( Attr: DWORD ) : DWORD;
begin
  Result := (Attr and $FFFF0000);
end; { Strip_fts1_attr }

{ --------------------------------------------------------- }
{ Strip_ext_attr                                            }
{ --------------------------------------------------------- }

function Strip_ext_attr( Attr: DWORD ) : DWORD;
begin
  Result := (Attr and $FFFF);
end; { Strip_ext_attr }

{ --------------------------------------------------------- }
{ Strip_netmail_attr                                        }
{ --------------------------------------------------------- }

const
  ECHOMAIL_ATTR_MASK = ATTR_LOCAL + ATTR_LOCK;

function Strip_netmail_attr( Attr: DWORD ) : DWORD;
begin
  Result := (Attr and ECHOMAIL_ATTR_MASK);
end; { Strip_netmail_attr }

{ --------------------------------------------------------- }
{ WriteStr                                                  }
{ --------------------------------------------------------- }

procedure WriteStr( var Stream: TStream; S: String; Term: Char );
begin
  if Length(S) > 0 then
    Stream.Write( S[1], Length(S) );
  Stream.Write( Term, 1 );
end; { WriteStr }

{ --------------------------------------------------------- }
{ ParseMsgDate                                              }
{ --------------------------------------------------------- }

//      1234567890123456789
//      dd mmm yy  hh:mm:ss

function ParseMsgDate( const S: String ) : TDateTime;

  function WhatMonth( const S: String ) : Integer;
  var
    j: Integer;
  begin
    for j := 1 to 12 do
      if cmpi( S, _MONTH_NAME[j] ) = 0 then
      begin
        Result := j;
        Exit;
      end;
    raise EConvertError.Create( 'Bad month' );
  end; { WhatMonth }

var
  Day, Month, Year, Hour, Min, Sec, MSec: SWORD;

begin
  if Length(S) <> 19 then
    raise EConvertError.Create( 'Bad date format' );
  Day   := StrToInt ( Copy(S, 1, 2) );
  Month := WhatMonth( Copy(S, 4, 3) );
  Year  := StrToInt ( Copy(S,  8, 2) );
  Hour  := StrToInt ( Copy(S, 12, 2) );
  Min   := StrToInt ( Copy(S, 15, 2) );
  Sec   := StrToInt ( Copy(S, 18, 2) );
  MSec  := 0;
  if Year < 30 then
    Inc(Year, 2000)
  else
    Inc(Year, 1900);
  Result := EncodeDate( Year, Month, Day ) + EncodeTime( Hour, Min, Sec, MSec );
end; { ParseMsgDate }

{ --------------------------------------------------------- }
{ MsgDateStr                                                }
{ --------------------------------------------------------- }

function MsgDateStr( D: TDateTime ) : String;
begin
  Result := FormatDateTime( 'dd mmm yy  hh:nn:ss', D );
end; { MsgDateStr }

{ --------------------------------------------------------- }
{ is_netmail                                                }
{ --------------------------------------------------------- }

function is_netmail( const echotag: String ) : Boolean;
begin
  Result := cmpi(echotag, _NETMAIL) = 0;
end; { is_netmail }

{ --------------------------------------------------------- }
{ Invalidate                                                }
{ --------------------------------------------------------- }

function Invalidate( S: String ) : String;
begin
  if Pos( _ORIGIN, S ) = 1 then
    S[2] := '+'
  else if Pos( _TEARLINE, S ) = 1 then
    S[2] := '+';
  Result := S;
end; { Invalidate }

{ --------------------------------------------------------- }
{ MakePrintable                                             }
{ --------------------------------------------------------- }

function MakePrintable( S: String ) : String;
var
  j: Integer;
begin
  Result := '';
  for j := 1 to Length(S) do
    if S[j] < ' ' then
      Result := Result + '^' + Chr( Ord(S[j]) + $40 )
    else
      Result := Result + S[j];
end; { MakePrintable }

////////////////////////////////////////////////////////////////
//
// TFtnMessage
//
////////////////////////////////////////////////////////////////

{ --------------------------------------------------------- }
{ Init                                                      }
{ --------------------------------------------------------- }

constructor TFtnMessage.Init;
begin
  inherited Init(100, 100);
  Path   := New( PStrings, Init(10, 10) );
  Seenby := New( PStrings, Init(10, 10) );
  Via    := New( PStrings, Init(10, 10) );
end; { Init }

{ --------------------------------------------------------- }
{ Done                                                      }
{ --------------------------------------------------------- }

destructor TFtnMessage.Done;
begin
  Destroy( Path );
  Destroy( Seenby );
  Destroy( Via );
  FreeStr( Origin );
  FreeStr( Tearline );
  inherited Done;
end; { Done }

{ --------------------------------------------------------- }
{ GetArea                                                   }
{ --------------------------------------------------------- }

function TFtnMessage.GetArea: String;
begin
  if Area = '' then
    Result := _NETMAIL
  else
    Result := Area;
end; { GetArea }

{ --------------------------------------------------------- }
{ GetFrom                                                   }
{ --------------------------------------------------------- }

function TFtnMessage.GetFrom: String;
begin
  Result := _From;
end; { GetFrom }

{ --------------------------------------------------------- }
{ GetTo                                                     }
{ --------------------------------------------------------- }

function TFtnMessage.GetTo: String;
begin
  Result := _To;
end; { GetTo }

{ --------------------------------------------------------- }
{ GetOrig                                                   }
{ --------------------------------------------------------- }

function TFtnMessage.GetOrig: TAddress;
begin
  Result := Orig;
end; { GetOrig }

{ --------------------------------------------------------- }
{ GetDest                                                   }
{ --------------------------------------------------------- }

function TFtnMessage.GetDest: TAddress;
begin
  Result := Dest;
end; { GetDest }

{ --------------------------------------------------------- }
{ GetSubj                                                   }
{ --------------------------------------------------------- }

function TFtnMessage.GetSubj: String;
begin
  Result := Subj;
end; { GetSubj }

{ --------------------------------------------------------- }
{ GetAttr                                                   }
{ --------------------------------------------------------- }

function TFtnMessage.GetAttr: DWORD;
begin
  Result := Attr;
end; { GetAttr }

{ --------------------------------------------------------- }
{ GetFlags                                                  }
{ --------------------------------------------------------- }

function TFtnMessage.GetFlags: String;
begin
  Result := Flags;
end; { GetFlags }

{ --------------------------------------------------------- }
{ GetDate                                                   }
{ --------------------------------------------------------- }

function TFtnMessage.GetDate: TDateTime;
begin
  Result := Date;
end; { GetDate }

{ --------------------------------------------------------- }
{ GetMsgID                                                  }
{ --------------------------------------------------------- }

function TFtnMessage.GetMsgID: String;
begin
  Result := MsgID;
end; { GetMsgID }

{ --------------------------------------------------------- }
{ GetReply                                                  }
{ --------------------------------------------------------- }

function TFtnMessage.GetReply: String;
begin
  Result := Reply;
end; { GetReply }

{ --------------------------------------------------------- }
{ HasOrigin                                                 }
{ --------------------------------------------------------- }

function TFtnMessage.HasOrigin: Boolean;
begin
  Result := Origin <> nil;
end; { HasOrigin }

{ --------------------------------------------------------- }
{ GetOrigin                                                 }
{ --------------------------------------------------------- }

function TFtnMessage.GetOrigin: String;
begin
  if Origin <> nil then
    Result := Origin^
  else
    Result := '';
end; { GetOrigin }

{ --------------------------------------------------------- }
{ MakeOrigin                                                }
{ --------------------------------------------------------- }

function TFtnMessage.MakeOrigin: String;
begin
  Result := _ORIGIN + GetOrigin + ' (' + Orig.MakeStr + ')';
end; { MakeOrigin }

{ --------------------------------------------------------- }
{ HasTearline                                               }
{ --------------------------------------------------------- }

function TFtnMessage.HasTearline: Boolean;
begin
  Result := Tearline <> nil;
end; { HasTearline }

{ --------------------------------------------------------- }
{ GetTearline                                               }
{ --------------------------------------------------------- }

function TFtnMessage.GetTearline: String;
begin
  if Tearline <> nil then
    Result := Tearline^
  else
    Result := '';
end; { GetTearline }

{ --------------------------------------------------------- }
{ MakeTearline                                              }
{ --------------------------------------------------------- }

function TFtnMessage.MakeTearline: String;
begin
  Result := _TEARLINE + GetTearline;
end; { MakeTearline }

{ --------------------------------------------------------- }
{ SetArea                                                   }
{ --------------------------------------------------------- }

procedure TFtnMessage.SetArea( const S: String );
begin
  if cmpi(S, _NETMAIL) = 0 then
    Area := ''
  else
    Area := S;
end; { SetArea }

{ --------------------------------------------------------- }
{ SetFrom                                                   }
{ --------------------------------------------------------- }

procedure TFtnMessage.SetFrom( const S: String );
begin
  _From := S;
end; { SetFrom }

{ --------------------------------------------------------- }
{ SetTo                                                     }
{ --------------------------------------------------------- }

procedure TFtnMessage.SetTo( const S: String );
begin
  _To := S;
end; { SetTo }

{ --------------------------------------------------------- }
{ SetOrig                                                   }
{ --------------------------------------------------------- }

procedure TFtnMessage.SetOrig( const A: TAddress );
begin
  if CompAddr( A, Orig ) = 0 then
    Orig := A
  else
  begin
    Orig := A;
    UpdateMsgID;
  end;
end; { SetOrig }

{ --------------------------------------------------------- }
{ SetDest                                                   }
{ --------------------------------------------------------- }

procedure TFtnMessage.SetDest( const A: TAddress );
begin
  Dest := A;
end; { SetDest }

{ --------------------------------------------------------- }
{ SetSubj                                                   }
{ --------------------------------------------------------- }

procedure TFtnMessage.SetSubj( const S: String );
begin
  Subj := S;
end; { SetSubj }

{ --------------------------------------------------------- }
{ SetAttr                                                   }
{ --------------------------------------------------------- }

procedure TFtnMessage.SetAttr( A: DWORD );
begin
  Attr := A;
end; { SetAttr }

{ --------------------------------------------------------- }
{ SetAttr                                                   }
{ --------------------------------------------------------- }

procedure TFtnMessage.SetFlags( const S: String );
begin
  Flags := S;
end; { SetFlags }

{ --------------------------------------------------------- }
{ SetDate                                                   }
{ --------------------------------------------------------- }

procedure TFtnMessage.SetDate( D: TDateTime );
begin
  Date := D;
end; { SetDate }

{ --------------------------------------------------------- }
{ SetReply                                                  }
{ --------------------------------------------------------- }

procedure TFtnMessage.SetReply( const S: String );
begin
  Reply := S;
end; { SetReply }

{ --------------------------------------------------------- }
{ SetOrigin                                                 }
{ --------------------------------------------------------- }

procedure TFtnMessage.SetOrigin( const S: String );
begin
  ReplaceStr( Origin, S );
end; { SetOrigin }

{ --------------------------------------------------------- }
{ DropOrigin                                                }
{ --------------------------------------------------------- }

procedure TFtnMessage.DropOrigin;
begin
  FreeStr( Origin );
end; { DropOrigin }

{ --------------------------------------------------------- }
{ SetTearline                                               }
{ --------------------------------------------------------- }

procedure TFtnMessage.SetTearline( const S: String );
begin
  ReplaceStr( Tearline, S );
end; { SetTearline }

{ -------------------------------------------------------- }
{ AddVia                                                   }
{ -------------------------------------------------------- }

procedure TFtnMessage.AddVia( const S: String );
begin
  Via^.Insert( AllocStr( ^A'Via ' + S ) );
end; { AddVia }

{ --------------------------------------------------------- }
{ Echomail                                                  }
{ --------------------------------------------------------- }

function TFtnMessage.Echomail: Boolean;
begin
  Result := Area <> '';
end; { Echomail }

{ --------------------------------------------------------- }
{ Kill                                                      }
{ --------------------------------------------------------- }

procedure TFtnMessage.Kill;
begin
  Rip := True;
end; { Kill }

{ --------------------------------------------------------- }
{ Deleted                                                   }
{ --------------------------------------------------------- }

function TFtnMessage.Deleted: Boolean;
begin
  Result := Rip;
end; { Deleted }

{ --------------------------------------------------------- }
{ BinSave                                                   }
{ --------------------------------------------------------- }

procedure TFtnMessage.BinSave( var bin: TStream );

  procedure WriteLine( P: VString ); far;
  begin
    WriteStr( bin, P^, _CR );
  end; { WriteLine }

var
  f: String;
  ExtAttr : DWORD;
  DestNode: TAddress;
  OrigNode: TAddress;

begin
  SaveBinHeader( bin );
  if Echomail then
    WriteStr( bin, 'AREA:' + Area, _CR );
  if MsgID <> '' then
    WriteStr( bin, ^A'MSGID: ' + MsgID, _CR );
  if Reply <> '' then
    WriteStr( bin, ^A'REPLY: ' + Reply, _CR );

  ExtAttr := Strip_fts1_attr( Attr );
  if Echomail then
    ExtAttr := Strip_netmail_attr( ExtAttr )
  else
  begin
    DestNode := Dest; DestNode.Point := 0;
    OrigNode := Orig; OrigNode.Point := 0;
    WriteStr( bin, ^A'INTL ' + DestNode.MakeStr + ' ' + OrigNode.MakeStr, _CR );
    if Orig.Point <> 0 then
      WriteStr( bin, ^A'FMPT ' + IntToStr(Orig.Point), _CR );
    if Dest.Point <> 0 then
      WriteStr( bin, ^A'TOPT ' + IntToStr(Dest.Point), _CR );
  end;

  f := FlagsStr( ExtAttr, Flags );
  if f <> '' then
    WriteStr( bin, ^A'FLAGS ' + f, _CR );

  ForEach( @WriteLine );

  if TearLine <> nil then
    WriteStr( bin, MakeTearline, _CR );
  if Origin <> nil then
    WriteStr( bin, MakeOrigin, _CR );

  if Echomail then
  begin
    Seenby^.ForEach( @WriteLine );
    Path^.ForEach( @WriteLine );
  end
  else
    Via^.ForEach( @WriteLine );

  bin.Write( _ZERO_BYTE, 1 );

end; { BinSave }

{ --------------------------------------------------------- }
{ TxtSave                                                   }
{ --------------------------------------------------------- }

procedure TFtnMessage.TxtSave( var Txt: Text );

  procedure WriteLine( V: VString ); far;
  begin
    Writeln( Txt, MakePrintable(V^) );
  end; { WriteLine }

var
  OrigNode: TAddress;
  DestNode: TAddress;
begin
  Writeln( Txt, CharStr('^', 75) );
  Writeln( Txt, '   Area:  ', GetArea );
  Writeln( Txt, '   From:  ', _From, ', ', Orig.MakeStr );
  if Echomail then
    Writeln( Txt, '     To:  ', _To )
  else
    Writeln( Txt, '     To:  ', _To, ', ', Dest.MakeStr );
  Writeln( Txt, '   Subj:  ', Subj );
  Writeln( Txt, '   Date:  ', MsgDateStr(Date) );
  if not Echomail then
    Writeln( Txt, '   Attr:  ', AttrStr(Attr) );
  Writeln( Txt, CharStr('^', 75) );
  if MsgID <> '' then
    Writeln( Txt, '^AMSGID: ', MsgID );
  if Reply <> '' then
    Writeln( Txt, '^AREPLY: ', Reply );
  if not Echomail then
  begin
    DestNode := Dest; DestNode.Point := 0;
    OrigNode := Orig; OrigNode.Point := 0;
    Writeln( Txt, '^AINTL ', DestNode.MakeStr, ' ', OrigNode.MakeStr );
    if Orig.Point <> 0 then
      Writeln( Txt, '^AFMPT ', Orig.Point );
    if Dest.Point <> 0 then
      Writeln( Txt, '^ATOPT ', Dest.Point );
  end;

  ForEach( @WriteLine );

  if HasTearline then
    Writeln( Txt, MakeTearline );

  if HasOrigin then
    Writeln( Txt, MakeOrigin );

  if Echomail then
  begin
    Seenby^.ForEach( @WriteLine );
    Path^.ForEach( @WriteLine );
  end
  else
    Via^.ForEach( @WriteLine );

end; { TxtSave }

{ --------------------------------------------------------- }
{ CheckKludge                                               }
{ --------------------------------------------------------- }

function TFtnMessage.CheckKludge( const S: String ) : Boolean;
var
  A    : TAddress;
  Key  : String;
  Val  : String;
  lbrac: Integer;
  rbrac: Integer;
  olen : Integer;

begin
  if S = '' then
  begin
    Result := False;
    Exit;
  end;
  Result := True;
  if S[1] = ^A then
  begin
    SplitPair( S, Key, Val );
    if Key = ^A'MSGID:' then
      MsgID := Val
    else if Key = ^A'INTL' then
    begin
      A.Assign( ExtractWord(1, Val, _BLANK) );
      if A.Valid then
      begin
        Dest.Zone := A.Zone;
        Dest.Net  := A.Net;
        Dest.Node := A.Node;
      end;
      A.Assign( ExtractWord(2, Val, _BLANK) );
      if A.Valid then
      begin
        Orig.Zone := A.Zone;
        Orig.Net  := A.Net;
        Orig.Node := A.Node;
      end;
    end
    else if Key = ^A'FMPT' then
    begin
      try
        Orig.Point := StrToInt(Val);
      except
        Orig.Point := 0;
      end
    end
    else if Key = ^A'TOPT' then
    begin
      try
        Dest.Point := StrToInt(Val);
      except
        Dest.Point := 0;
      end
    end
    else if Key = ^A'REPLY:' then
      Reply := Val
    else if Key = ^A'FLAGS' then
      ParseFlags( Val )
    else if (Key = ^A'Via') or (Key = ^A'Recd') then
      Via^.Insert( AllocStr(S) )
    else if Key = ^A'PATH:' then
      Path^.Insert( AllocStr(S) )
    else if Key = 'SEEN-BY:' then
      SeenBy^.Insert( AllocStr(S) )
    else
      Result := False;
  end
  else
  begin
    if Pos(_ORIGIN, S) = 1 then
    begin
      rbrac := Length(S);
      while (rbrac > 0) and (S[rbrac] <> ')') do Dec(rbrac);
      lbrac := rbrac;
      while (lbrac > 0) and (S[lbrac] <> '(') do Dec(lbrac);

      olen  := lbrac - Length(_ORIGIN) - 1;
      if olen > 0 then
        ReplaceStr( Origin, Copy( S, Length(_ORIGIN) + 1, olen ) )
      else
        ReplaceStr( Origin, '' );
      if Echomail and (lbrac + 1 < rbrac ) then
      begin
        lbrac := rbrac - 1;
        while (S[lbrac] <> '(') and (S[lbrac] <> ' ') do Dec(lbrac);
        A.Assign( Copy(S, lbrac+1, rbrac - lbrac -  1 ) );
        if A.Valid then
          Orig := A;
      end
    end
    else if Pos(_TEARLINE, S ) = 1 then
      ReplaceStr( Tearline, Copy( S, Length(_TEARLINE) + 1, Length(S) ))
    else
      Result := False;
  end;
end; { CheckKludge }

{ --------------------------------------------------------- }
{ ParseFlags                                                }
{ --------------------------------------------------------- }

procedure TFtnMessage.ParseFlags( const S: String );
var
  j: Integer;
  n: Integer;
  a: DWORD;
  f: String;
begin
  Flags := '';
  n := WordCount( S, _BLANK );
  for j := 1 to n do
  begin
    f := ExtractWord( j, S, _BLANK );
    a := WhatsAttr( f );
    if a <> 0 then
      SetBit( Attr, a, True )
    else
      Flags := Flags + f + ' ';
  end;
  Flags := TrimRight( Flags );
end; { ParseFlags }

{ --------------------------------------------------------- }
{ UpdateMsgID                                               }
{ --------------------------------------------------------- }

const
  LastID: DWORD = 0;

procedure TFtnMessage.UpdateMsgID;
begin
  MsgID := Orig.MakeStr + ' ' + IntToHex( CurrentUnixTime shl 3 + LastID, 8 );
  Inc( LastID );
end; { UpdateMsgID }

{ --------------------------------------------------------- }
{ SaveBinHeader                                             }
{ --------------------------------------------------------- }

procedure TFtnMessage.SaveBinHeader;
begin
  Abstract;
end; { SaveBinHeader }

{ --------------------------------------------------------- }
{ Clone                                                     }
{ --------------------------------------------------------- }

procedure TFtnMessage.Clone( Source: PFtnMessage );

  procedure CopyStrings( V1, V2: PCollection );
    procedure MakeCopy( V: VString ); far;
    begin
      V2^.Insert( AllocStr(V^) );
    end; { MakeCopy }
  begin
    V1.ForEach( @MakeCopy );
  end; { CopyStrings }

begin
  Area  := Source^.Area;
  Orig  := Source^.Orig;
  Dest  := Source^.Dest;
  _From := Source^._From;
  _To   := Source^._To;
  Subj  := Source^.Subj;
  Date  := Source^.Date;
  Attr  := Source^.Attr;
  Flags := Source^.Flags;
  MsgID := Source^.MsgID;
  Reply := Source^.Reply;
  Rip   := Source^.Rip;

  if Source^.Tearline <> nil then
    Tearline := AllocStr( Source^.Tearline^ );

  if Source^.Origin <> nil then
    Origin := AllocStr( Source^.Origin^ );

  CopyStrings( Source, @Self );
  CopyStrings( Source^.Seenby, Seenby );
  CopyStrings( Source^.Path, Path );
  CopyStrings( Source^.Via, Via );

end; { Clone }

////////////////////////////////////////////////////////////////
//
// NetmailMessage
//
////////////////////////////////////////////////////////////////

type
  PNetmailHeader = ^TNetmailHeader;
  TNetmailHeader = packed record
    _From: array [0..35] of Char;
    _To  : array [0..35] of Char;
    Subj : array [0..71] of Char;
    Date : array [0..19] of Char;
    TimesRead: SWORD;
    DestNode : SWORD;
    OrigNode : SWORD;
    Cost     : SWORD;
    OrigNet  : SWORD;
    DestNet  : SWORD;
    DestZone : SWORD;
    OrigZone : SWORD;
    DestPoint: SWORD;
    OrigPoint: SWORD;
    ReplyTo  : SWORD;
    Attr     : SWORD;
    NextReply: SWORD;
  end; { TNetmailHeader }

{$IFDEF USE_FILE_MAPPING}

{ --------------------------------------------------------- }
{ Load                                                      }
{ --------------------------------------------------------- }

procedure TNetmailMessage.Load( FileName: String );
var
  B: array [0..79] of Char;
  S: String;
  A: TAddress;
  P: PChar;
  Lim: PChar;
  hFile   : THandle;
  hMapping: THandle;
  Header  : PNetmailHeader;

  function MsgGetStr( var S: String ) : Boolean;
  var
    N: Integer;
    Start: PChar;
  begin
    N := 0;
    Start := P;
    while P < Lim do
    begin
      if P[0] = _CR then
      begin
        Inc(P);
        SetLength( S, N );
        if N > 0 then
          Move( Start[0], S[1], N );
        if P[0] = _LF then Inc(P);
        Result := True;
        Exit;
      end;
      if P[0] = #0 then
      begin
        SetLength( S, N );
        if N > 0 then
          Move( Start[0], S[1], N );
        Result := False;
        Exit;
      end;
      Inc(P);
      Inc(N);
    end;
    Result := False;
  end; { MsgGetStr }

begin
  MsgName  := FileName;
  FileName := FileName + #0;

  hFile := CreateFile( @FileName[1], GENERIC_READ, FILE_SHARE_READ,
                       nil, OPEN_EXISTING, FILE_FLAG_SEQUENTIAL_SCAN, 0 );

  if hFile = 0 then
    raise Exception.Create( 'CreateFile failed' );

  hMapping := CreateFileMapping( hFile, nil, PAGE_READONLY, 0, 0, nil );

  if hMapping = 0 then
  begin
    CloseHandle( hFile );
    raise Exception.Create( 'CreateFileMapping failed' );
  end;

  Header := MapViewOfFile( hMapping, FILE_MAP_READ, 0, 0, 0 );

  if Header = nil then
  begin
    CloseHandle( hMapping );
    CloseHandle( hFile );
    raise Exception.Create( 'MapViewOfFile failed' );
  end;

  _From := StrPas( StrLCopy(B, Header^._From, 36) );
  _To   := StrPas( StrLCopy(B, Header^._To,   36) );
  Subj  := StrPas( StrLCopy(B, Header^.Subj,  72) );
  Attr  := Header^.Attr;
  Area  := '';

  try
    Date := ParseMsgDate( StrPas(StrLCopy(B, Header^.Date, 20) ) );
  except
    Date := SysUtils.Now;
  end;

  with Header^ do
  begin
    Orig.Merge( 0, OrigNet, OrigNode, 0 );
    Dest.Merge( 0, DestNet, DestNode, 0 );
  end;

  P   := PChar(Header) + SizeOf(TNetmailHeader);
  Lim := PChar(Header) + GetFileSize( hFile, nil );

  while MsgGetStr( S ) do
    if not CheckKludge( S ) then
      Insert( AllocStr(S) );

  if Orig.Zone = 0 then
  begin
    A.Assign( ExtractWord(1, MsgID, _BLANK) );
    if A.Valid then
      Orig.Zone := A.Zone;
  end;

  if Dest.Zone = 0 then
    Dest.Zone := Orig.Zone;

  UnmapViewOfFile( Header );
  CloseHandle( hMapping );
  CloseHandle( hFile );

end; { Load }

{$ENDIF}

{ --------------------------------------------------------- }
{ SaveBinHeader                                             }
{ --------------------------------------------------------- }

procedure TNetmailMessage.SaveBinHeader( var bin: TStream );
var
  Header: TNetmailHeader;
begin
  FillChar( Header, SizeOf(Header), 0 );
  StrSet( Header._From, _From, 36 );
  StrSet( Header._To, _To, 36 );
  StrSet( Header.Subj, Subj, 72 );
  StrSet( Header.Date, MsgDateStr(Date), 20 );
  Header.OrigZone  := Orig.Zone;
  Header.OrigNet   := Orig.Net;
  Header.OrigNode  := Orig.Node;
  Header.OrigPoint := Orig.Point;
  Header.DestZone  := Dest.Zone;
  Header.DestNet   := Dest.Net;
  Header.DestNode  := Dest.Node;
  Header.DestPoint := Dest.Point;
  Header.Attr      := Strip_ext_Attr( Attr );
  bin.Write( Header, SizeOf(Header) );
end; { SaveBinHeader }

{ --------------------------------------------------------- }
{ Save                                                      }
{ --------------------------------------------------------- }

procedure TNetmailMessage.Save;
var
  bin: TBufStream;
begin
  if Deleted then Exit;
  bin.Init( MsgName, stCreate, STREAM_BUFFER_SIZE );
  if bin.Status = stOk then
    BinSave( bin );
  bin.Done;
end; { Save }

{ --------------------------------------------------------- }
{ SaveAs                                                    }
{ --------------------------------------------------------- }

procedure TNetmailMessage.SaveAs( const FileName: String );
begin
  MsgName := FileName;
  Save;
end; { SaveAs }

{ --------------------------------------------------------- }
{ Kill                                                      }
{ --------------------------------------------------------- }

procedure TNetmailMessage.Kill;
begin
  inherited Kill;
  SysUtils.DeleteFile( MsgName );
end; { Kill }

{ --------------------------------------------------------- }
{ NewMsgFileName                                            }
{ --------------------------------------------------------- }

var
  LastMsgNum : Integer;

function NewMsgFileName( const Path: String ) : String;
var
  j  : Integer;
  Top: Integer;
  SR : TSearchRec;
  DosError: Integer;
begin
  Top := 0;
  DosError := SysUtils.FindFirst( AtPath('*.MSG', Path), faArchive or faReadOnly, SR );
  while DosError = 0 do
  begin
    try
      j := StrToInt( ChangeFileExt(SR.Name, '') );
      if j > Top then
        Top := j;
    except
      { ...nothing...}
    end;
    DosError := SysUtils.FindNext( SR );
  end;
  SysUtils.FindClose( SR );
  LastMsgNum  := Succ(Top);
  Result := AtPath( IntToStr(LastMsgNum) + '.MSG', Path );
end; { NewMsgFileName }

{ --------------------------------------------------------- }
{ NextMsgFileName                                           }
{ --------------------------------------------------------- }

function NextMsgFileName( const Path: String ) : String;
begin
  Inc( LastMsgNum );
  Result := AtPath( IntToStr(LastMsgNum) + '.MSG', Path );
end; { NextMsgFileName }


////////////////////////////////////////////////////////////////
//
// PackedMessage
//
////////////////////////////////////////////////////////////////

type
  PPackedMessageHeader = ^TPackedMessageHeader;
  TPackedMessageHeader = packed record
    Version : SWORD;
    OrigNode: SWORD;
    DestNode: SWORD;
    OrigNet : SWORD;
    DestNet : SWORD;
    Attr    : SWORD;
    Cost    : SWORD;
  end; { TPackedMessageHeader }

{$IFDEF USE_FILE_MAPPING}

{ --------------------------------------------------------- }
{ Load                                                      }
{ --------------------------------------------------------- }

procedure TPackedMessage.Load( var Pkt: PChar; PktOrig, PktDest: TAddress );

 function PktGetStr( term: Char ) : String;
 var
   p: PChar;
 begin
   p := pkt;
   pkt := StrScan( pkt, term );
   SetLength( Result, pkt - p );
   if pkt > p then
     Move( p[0], Result[1], pkt - p );
   Inc( pkt );
   if pkt[0] = _LF then
     Inc( pkt )
 end; { PktGetStr }

var
  S: String;
  A: TAddress;
  Header: PPackedMessageHeader;

begin
  Header    := PPackedMessageHeader( pkt );
  Attr      := Header^.Attr;

  with Header^ do
  begin
    Orig.Merge( 0, OrigNet, OrigNode, 0 );
    Dest.Merge( 0, DestNet, DestNode, 0 );
  end;

  Inc( pkt, SizeOf(TPackedMessageHeader) );
  try
    Date := ParseMsgDate( PktGetStr( #0 ) );
  except
    Date := Now;
  end;
  _To   := PktGetStr( #0 );
  _From := PktGetStr( #0 );
  Subj  := PktGetStr( #0 );
  if StrLIComp( pkt, 'AREA:', 5 ) = 0 then
    Area := Trim( Copy( PktGetStr( _CR ), 6, 255 ) );

  while pkt[0] <> #0 do
  begin
    S := PktGetStr( _CR );
    if not CheckKludge( S ) then
      Insert( AllocStr(S) );
  end;

  Inc( pkt );

  if Orig.Zone = 0 then
    Orig.Zone := PktOrig.Zone;

  if Dest.Zone = 0 then
    Dest.Zone := PktDest.Zone;

end; { Load }

{$ENDIF}

{ --------------------------------------------------------- }
{ SaveBinHeader                                             }
{ --------------------------------------------------------- }

procedure TPackedMessage.SaveBinHeader( var bin: TStream );
var
  Header: TPackedMessageHeader;
begin
  FillChar( Header, SizeOf(Header), #0 );
  Header.Version  := 2;
  Header.OrigNode := Orig.Node;
  Header.DestNode := Dest.Node;
  Header.OrigNet  := Orig.Net;
  Header.DestNet  := Dest.Net;
  if EchoMail then
    Header.Attr := Strip_ext_attr( Strip_netmail_attr(Attr) )
  else
    Header.Attr := Strip_ext_attr( Attr );
  bin.Write( Header, SizeOf(Header) );
  WriteStr( bin, MsgDateStr(Date), #0 );
  WriteStr( bin, _To,   #0 );
  WriteStr( bin, _From, #0 );
  WriteStr( bin, Subj,  #0 );
end; { SaveBinHeader }

////////////////////////////////////////////////////////////////
//
// Packet
//
////////////////////////////////////////////////////////////////

type
  PPacketHeader = ^TPacketHeader;
  TPacketHeader = packed record
    OrigNode: SWORD;
    DestNode: SWORD;
    Year    : SWORD;
    Month   : SWORD;
    Day     : SWORD;
    Hour    : SWORD;
    Min     : SWORD;
    Sec     : SWORD;
    Baud    : SWORD;
    PktType : SWORD;
    OrigNet : SWORD;
    DestNet : SWORD;
    ProdCode: BYTE;
    SerialNo: BYTE;
    Password: array [0..7] of Char;
    OrigZone: SWORD;
    DestZone: SWORD;
    FillArea: array [0..7] of Char;
    OrigZoneGate: SWORD;
    DestZoneGate: SWORD;
    OrigPoint: SWORD;
    DestPoint: SWORD;
    Mark: array [0..3] of Char;
  end; { TPacketHeader }

{ --------------------------------------------------------- }
{ Init                                                      }
{ --------------------------------------------------------- }

constructor TPacket.Init;
begin
  inherited Init(100, 100);
end; { Init }

{$IFDEF USE_FILE_MAPPING}

{ --------------------------------------------------------- }
{ Load                                                      }
{ --------------------------------------------------------- }

procedure TPacket.Load( FileName: String );
var
  B: array [0..10] of Char;
  Pkt: PChar;
  Msg: PPackedMessage;
  hFile: THandle;
  hMapping: THandle;
  Header  : PPacketHeader;
begin
  PktName  := FileName;
  FileName := FileName + #0;

  hFile := CreateFile( @FileName[1], GENERIC_READ, FILE_SHARE_READ,
                       nil, OPEN_EXISTING, FILE_FLAG_SEQUENTIAL_SCAN, 0 );

  if hFile = 0 then
    raise Exception.Create( 'CreateFile failed' );

  hMapping := CreateFileMapping( hFile, nil, PAGE_READONLY, 0, 0, nil );

  if hMapping = 0 then
  begin
    CloseHandle( hFile );
    raise Exception.Create( 'CreateFileMapping failed' );
  end;

  Header := MapViewOfFile( hMapping, FILE_MAP_READ, 0, 0, 0 );

  if Header = nil then
  begin
    CloseHandle( hMapping );
    CloseHandle( hFile );
    raise Exception.Create( 'MapViewOfFile failed' );
  end;

  with Header^ do
  begin
    Orig.Merge( OrigZone, OrigNet, OrigNode, OrigPoint );
    Dest.Merge( DestZone, DestNet, DestNode, DestPoint );
    Date := EncodeDate( Year, Succ(Month), Day ) +
            EncodeTime(Hour, Min, Sec, 0 );
  end;

  FillChar( B, SizeOf(B), 0 );
  Password := StrPas( StrLCopy( B, Header^.Password, 8 ) );

  pkt := PChar(Header) + SizeOf(TPacketHeader);

  try
    while pkt[0] <> #0 do
    begin
      Msg := New( PPackedMessage, Init );
      Msg^.Load( Pkt, Orig, Dest );
      Insert( Msg );
      Msg := nil;
    end;
  finally
    Destroy( Msg );                    // if exception has been raised
    UnmapViewOfFile( Header );
    CloseHandle( hMapping );
    CloseHandle( hFile );
  end;

end; { Load }

{$ENDIF}

{ --------------------------------------------------------- }
{ Save                                                      }
{ --------------------------------------------------------- }

procedure TPacket.Save;
var
  bin: TBufStream;
  Header: TPacketHeader;
  Dummy : SWORD;

  procedure SaveMsg( M: PPackedMessage ); far;
  begin
    if not M^.Deleted then
      M^.BinSave( bin );
  end; { SaveMsg }

  function Alive( M: PPackedMessage ) : Boolean; far;
  begin
    Result := not M^.Deleted;
  end; { Alive }

begin
  if FirstThat( @Alive ) = nil then
  begin
    SysUtils.DeleteFile( PktName );
    Exit;
  end;

  bin.Init( PktName, stCreate, STREAM_BUFFER_SIZE );
  if bin.Status <> stOk then
  begin
    bin.Done;
    raise EInOutError.Create( 'Could not write packet "' + PktName + '"' );
  end;

  FillChar( Header, SizeOf(Header), 0 );
  with Header do
  begin
    PktType   := 2;
    OrigZone  := Orig.Zone;
    OrigNet   := Orig.Net;
    OrigNode  := Orig.Node;
    OrigPoint := Orig.Point;
    DestZone  := Dest.Zone;
    DestNet   := Dest.Net;
    DestNode  := Dest.Node;
    DestPoint := Dest.Point;
    OrigZoneGate := Orig.Zone;
    DestZoneGate := Dest.Zone;
    DecodeDate( Date, Year, Month, Day );
    Dec(Month);
    DecodeTime( Date, Hour, Min, Sec, Dummy );
  end;
  StrSet( Header.Password, Password, 8 );

  bin.Write( Header, SizeOf(Header) );
  ForEach( @SaveMsg );
  bin.Write( _ZERO_WORD, 2 );
  bin.Done;
end; { Save }

{ --------------------------------------------------------- }
{ SaveAs                                                    }
{ --------------------------------------------------------- }

procedure TPacket.SaveAs( const FileName: String );
begin
  PktName := FileName;
  Save;
end; { SaveAs }

{ --------------------------------------------------------- }
{ SetOrig                                                   }
{ --------------------------------------------------------- }

procedure TPacket.SetOrig( const A: TAddress );
begin
  Orig := A;
end; { SetOrig }

{ --------------------------------------------------------- }
{ SetDest                                                   }
{ --------------------------------------------------------- }

procedure TPacket.SetDest( const A: TAddress );
begin
  Dest := A;
end; { SetDest }

{ --------------------------------------------------------- }
{ SetDate                                                   }
{ --------------------------------------------------------- }

procedure TPacket.SetDate( D: TDateTime );
begin
  Date := D;
end; { SetDate }

{ --------------------------------------------------------- }
{ NewPktFileName                                            }
{ --------------------------------------------------------- }

function NewPktFileName( const Path: String ) : String;
var
  hash: Longint;
begin
  hash := (CurrentUnixTime shl 8) or
{$IFDEF VIRTUALPASCAL}
          (SysSysMsCount and $FFFF);
{$ELSE}
          (GetTickCount and $FFFF);
{$ENDIF}
  Result := AtPath( IntToHex(hash, 8) + '.PKT', Path );
end; { MakePktName }


end.

